"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.convertBashToWin = void 0;
var tslib_1 = require("tslib");
// patch for irrelevant node dependency of bash-parser:
if (typeof window === 'undefined') {
    // @ts-ignore
    var window_1 = { process: { env: { NODE_NEV: 'mock' } } }; // lgtm [js/unused-local-variable]
    // @ts-ignore
}
else if (!window.process) {
    // @ts-ignore
    window.process = { env: { NODE_NEV: 'mock' } };
}
var convert_paths_1 = require("./convert-paths");
var types_1 = require("./types");
var bash_parser_1 = tslib_1.__importDefault(require("bash-parser"));
var rm_handler_1 = require("./rm-handler");
var cp_handler_1 = require("./cp-handler");
var ConvertBash = /** @class */ (function () {
    function ConvertBash() {
        var _this = this;
        this.rmHandler = new rm_handler_1.RmHandler(function (c) { return _this.convertCommand(c); });
        this.cpHandler = new cp_handler_1.CpHandler(function (c) { return _this.convertCommand(c); });
        this.userDefinedFunctionNames = new Set();
        this.userDefinedFunctions = [];
        this.delayedExpansionActive = false;
        this.preStatements = [];
        this.interpolationCounter = 0;
    }
    ConvertBash.prototype.convertCommand = function (command) {
        var result = this.convertCommandCore(command);
        var preStatements = types_1.isStatement(command) ? this.drainPreStatements() : '';
        return preStatements + result;
    };
    ConvertBash.prototype.convertCommandCore = function (command) {
        var _this = this;
        console.log('convertCommand', command);
        switch (command.type) {
            case 'Command':
                if (command.prefix && command.prefix.length && (!command.name || !command.name.text)) { // simple variable assignment
                    return command.prefix.map(function (c) { return _this.convertCommand(c); }).join('\n');
                }
                if (command.name && command.name.text) {
                    if (this.userDefinedFunctionNames.has(command.name.text)) {
                        var params = command.suffix && command.suffix.length ? ' ' + command.suffix.map(function (x) { return _this.convertCommand(x); }).join(' , ') : '';
                        return "CALL :" + command.name.text + params;
                    }
                    var suffix = command.suffix ? " " + command.suffix.map(function (c) { return _this.convertCommand(c); }).join(' ') : '';
                    switch (command.name.text) {
                        case 'set':
                            if (command.suffix && command.suffix.length === 1 && command.suffix[0].type === 'Word' && command.suffix[0].text === '-e') {
                                console.log('skipping "set -e"');
                                return '';
                            }
                            else {
                                return "" + command.name.text + suffix;
                            }
                        case 'rm':
                            return this.rmHandler.handle(command);
                        case 'cp':
                            return this.cpHandler.handle(command);
                        default:
                            return "" + command.name.text + suffix;
                    }
                }
                return 'unknown command: ' + JSON.stringify(command);
            case 'Function':
                // NOTE: bash: definition before call. batch: call before definition
                // --> append function definitions at the end
                this.userDefinedFunctionNames.add(command.name.text);
                this.userDefinedFunctions.push(command);
                return '';
            case 'Word':
                var expandedWord = this.performExpansions(command.text, command.expansion);
                var textWord = convert_paths_1.convertPaths(expandedWord);
                if (textWord.startsWith('"')) {
                    /* Keep textWord as it is. */
                }
                else if (['=='].includes(textWord)) {
                    /* Keep textWord as it is. */
                }
                else if (['!=', '-ne'].includes(textWord)) {
                    textWord = 'NEQ';
                }
                else if (['-eq'].includes(textWord)) {
                    textWord = 'EQU';
                }
                else if (['-lt'].includes(textWord)) {
                    textWord = 'LSS';
                }
                else if (['-le'].includes(textWord)) {
                    textWord = 'LEQ';
                }
                else if (['-gt'].includes(textWord)) {
                    textWord = 'GTR';
                }
                else if (['-ge'].includes(textWord)) {
                    textWord = 'GEQ';
                }
                else if (['!'].includes(textWord)) {
                    textWord = 'NOT';
                }
                else {
                    textWord = "\"" + textWord + "\"";
                }
                return textWord;
            case 'AssignmentWord':
                var expandedAssignmentWord = this.performExpansions(command.text, command.expansion);
                var textAssignmentWord = convert_paths_1.convertPaths(expandedAssignmentWord);
                var _a = textAssignmentWord.split('=', 2), variableName = _a[0], variableValue = _a[1];
                var unquotedValue = variableValue.replace(/^"(.*)"$/, '$1');
                return "SET \"" + variableName + "=" + unquotedValue + "\"";
            case 'LogicalExpression':
                switch (command.op) {
                    case 'and':
                        return (this.convertCommand(command.left)) + " && " + (this.convertCommand(command.right));
                    case 'or':
                        return (this.convertCommand(command.left)) + " || " + (this.convertCommand(command.right));
                    default:
                        return "REM UNKNOWN operand \"" + command.op + "\" in: " + JSON.stringify(command);
                }
            case 'CompoundList':
                return command.commands.map(function (c) { return _this.convertCommand(c); }).join('\n');
            case 'If':
                // note: AND/OR is not supported with batch IF (https://stackoverflow.com/a/2143203/2544163)
                var condition = this.convertCommand(command.clause.commands[0]).replace(/^\[ |^"\[" | "]"$/g, '');
                var elseBranch = command.else ? " ELSE (\n" + this.indent(this.convertCommand(command.else)) + "\n)" : '';
                return "IF " + condition + " (\n" + this.indent(this.convertCommand(command.then)) + "\n)" + elseBranch;
            case 'Case':
                var caseStatement_1 = this.convertCommand(command.clause);
                return command.cases.map(function (c, i) {
                    var pattern = c.pattern[0]; // this is a list for unclear reason..
                    // simple heuristic: '*' is default case:
                    if (pattern.text === '*') {
                        return " ELSE (\n" + _this.indent(_this.convertCommand(c.body)) + "\n)";
                    }
                    var caseCondition = caseStatement_1 + "==" + _this.convertCommand(pattern);
                    var prefix = i === 0 ? 'IF' : ' ELSE IF';
                    return prefix + " " + caseCondition + " (\n" + _this.indent(_this.convertCommand(c.body)) + "\n)";
                }).join('');
        }
        return 'REM UNKNOWN: ' + JSON.stringify(command);
    };
    ConvertBash.prototype.getFunctionDefinitions = function () {
        var _this = this;
        if (!this.userDefinedFunctions.length) {
            return '';
        }
        return "\n\nEXIT /B %ERRORLEVEL%\n\n" + this.userDefinedFunctions.map(function (f) {
            var innerCommands = f.body.commands.map(function (c) { return _this.convertCommand(c); }).join('\n');
            return ":" + f.name.text + "\n" + innerCommands + "\nEXIT /B 0\n";
        }).join('\n');
    };
    ConvertBash.prototype.performExpansions = function (text, expansions) {
        // currently assumes only ParameterExpansions (TODO ArithmeticExpansion)
        var result = text || '';
        var sortedExpansions = tslib_1.__spreadArray([], (expansions || []));
        sortedExpansions.sort(function (a, b) { return a.loc.start > b.loc.start ? -1 : 1; });
        for (var _i = 0, sortedExpansions_1 = sortedExpansions; _i < sortedExpansions_1.length; _i++) {
            var expansion = sortedExpansions_1[_i];
            switch (expansion.type) {
                case 'CommandExpansion':
                    this.delayedExpansionActive = true;
                    var interpolationVar = "_INTERPOLATION_" + this.interpolationCounter++;
                    this.preStatements.push("SET " + interpolationVar + "=");
                    this.preStatements.push("FOR /f \"delims=\" %%a in ('" + expansion.command + "') DO (SET \"" + interpolationVar + "=!" + interpolationVar + "! %%a\")");
                    result = result.substring(0, expansion.loc.start) + "!" + interpolationVar + ":~1!" + result.substring(expansion.loc.end + 1);
                    break;
                case 'ParameterExpansion':
                    if (expansion.loc.start >= 0 && expansion.loc.end >= 0) { // in test 'should transform complete example' the loc is negative (and there is no meaningful expansion at that point)
                        // expand function parameters such as `$1` (-> `%~1`) different to regular variables `$MY`(-> `%MY%` or `!MY!` if delayed expansion is active):
                        var expandedValue = /^\d+$/.test("" + expansion.parameter) ? "%~" + expansion.parameter : (this.delayedExpansionActive ? "!" + expansion.parameter + "!" : "%" + expansion.parameter + "%");
                        result = "" + result.substring(0, expansion.loc.start) + expandedValue + result.substring(expansion.loc.end + 1);
                    }
                    break;
            }
        }
        return result;
    };
    ConvertBash.prototype.indent = function (s) {
        return s.split('\n').map(function (line) { return "  " + line; }).join('\n');
    };
    ConvertBash.prototype.delayedExpansion = function () {
        return this.delayedExpansionActive;
    };
    ConvertBash.prototype.drainPreStatements = function () {
        var result = this.preStatements.join('\n');
        this.preStatements = [];
        return result + (result ? '\n' : '');
    };
    return ConvertBash;
}());
function preprocess(script) {
    return script.replace(/(^|\n)\s*function /g, '$1')
        .replace(/\n(?=\n)/g, '\n_BLANK_LINE_="x"')
        .replace(/(^|\n)(?!#!)\s*#(.*)/g, '$1_COMMENT_="$2"');
}
function postprocess(script) {
    return script.replace(/SET "_COMMENT_= ?(.*)"/g, 'REM $1')
        .replace(/SET "_BLANK_LINE_=x"/g, '');
}
function convertBashToWin(script) {
    var preprocessedScript = preprocess(script);
    var ast = bash_parser_1.default(preprocessedScript, { mode: 'bash' });
    var converter = new ConvertBash();
    var convertedCommands = ast.commands
        .map(function (c) { return converter.convertCommand(c); })
        .filter(function (c) { return !!c; }) // filter empty commands
        .join('\n');
    var functionDefinitions = converter.getFunctionDefinitions();
    return '@echo off' +
        (converter.delayedExpansion() ? '\nsetlocal EnableDelayedExpansion' : '') +
        '\n' +
        postprocess(convertedCommands +
            functionDefinitions);
}
exports.convertBashToWin = convertBashToWin;
